
from __future__ import division, print_function

from itertools import chain
from math import ceil, log10
from multiprocessing import cpu_count, current_process
from os import close, getenv, remove
from os.path import abspath, exists
from re import finditer
from subprocess import PIPE, Popen
from sys import stderr
from tempfile import mkstemp
from textwrap import dedent
from warnings import warn

from fakemp import farmout, farmworker

from ._hyphyinterface import HyphyInterface, escape, tohyphy


__all__ = ['HyphyMap', 'mpi_node_count']


def _mpienvvar():
    try:
        node_count = int(getenv('MPI', '0'))
    except ValueError:
        warn('the MPI env var should specify the desired number of nodes, defaulting to 0')
        node_count = 0
    return node_count


def _runhyphympi(cmds, node_count=None):
    if node_count is None:
        node_count = _mpienvvar()
    fd, filename = mkstemp(); close(fd)
    try:
        with open(filename, 'w') as fh:
            fh.write('MESSAGE_LOGGING = 0;\n')
            fh.write(cmds)
        p = Popen(
            ['HYPHYMPI', filename],
            env={
                'NP': '%d' % node_count,
                'PATH': getenv('PATH', '/usr/local/bin:/usr/bin:/bin')
            },
            stderr=PIPE, stdout=PIPE
        )
        pout_b, perr_b = p.communicate()
        pout = pout_b.decode('utf-8').strip()
        perr = perr_b.decode('utf-8').strip()
        retcode = p.returncode
    except OSError as e:
        print(e.strerror, file=stderr)
        retcode = 1
        pout, perr = '', e.strerror
    finally:
        # make sure to clean up
        if exists(filename):
            remove(filename)

    return (retcode, pout, perr)


def mpi_node_count():
    node_count = _mpienvvar()
    if node_count > 1:
        try:
            # use MPI_NODE_COUNT - 1 so we don't count the root process
            cmds = 'fprintf( stdout, "" + (MPI_NODE_COUNT - 1) );'
            retcode, pout, perr = _runhyphympi(cmds)
            node_count = int(pout)
        except ValueError:
            node_count = 0
    else:
        node_count = 0
    return node_count


def _quicksize(value):
    return int(log10(max(1, value)))


def _jobopts(argslist):
    if isinstance(argslist, (list, tuple)):
        # both key and value must be strings or Hyphy bugs out
        return '_jobopts = {};\n' + (
            '\n'.join('_jobopts[ %d ] = { %s };' % (
                i,
                (',\n' + (' ' * (18 + _quicksize(i)) )).join('%s: %s' % (
                    # this shit is required because keys are string-sorted, so "10" comes before "2"
                    '"%s%d"' % ('0' * (_quicksize(len(args) - 1) - _quicksize(j)), j),
                    '"%s"' % repr(v) if isinstance(v, (int, float)) else escape(v)
                ) for j, v in enumerate(args)) if args is not None else ''
            ) for i, args in enumerate(argslist))
        )
    else:
        raise ValueError('Invalid argslist supplied: "%s"' % repr(argslist))


def _globalvars(varsdict):
    return '\n'.join(tohyphy(k, v) for k, v in varsdict.items())


def _thyphyexprs(numjobs):
    return '\n'.join(dedent('''\
        if ( key == "val%(jobid)d" ) {
                return _jobvals[ %(jobid)d ];
        }
    $''') % {
        'jobid': i
    } for i in range(numjobs)).lstrip().rstrip('\n$')


def _jobdispatch(batchfile, retvar, globals, argslist, quiet=True):
    numjobs = len(argslist)
    iface = HyphyInterface()
    cmds = dedent('''\
    %(globals)s
    _job = 0;
    %(jobopts)s
    _jobvals = {};
    for ( _job = 0; _job < %(numjobs)d; _job += 1 ) {
        ExecuteAFile( "%(batchfile)s", _jobopts[ _job ] );
        _jobvals[ _job ] = %(retvar)s;
    }
    function _THyPhyAskFor( key ) {
        %(thyphyexprs)s
        return "_THyPhy_NOT_HANDLED_";
    }''') % {
        'batchfile': batchfile,
        'globals': globals,
        'jobopts': _jobopts(argslist),
        'numjobs': numjobs,
        'retvar': retvar,
        'thyphyexprs': _thyphyexprs(numjobs),
    }
    iface.queuecmd(cmds)
    iface.runqueue()

    if not quiet:
        if iface.stdout != '':
            print(iface.stdout, file=stderr)
        if iface.warnings != '':
            print(iface.warnings, file=stderr)

    if iface.stderr != '':
        raise RuntimeError(iface.stderr)

    return [ iface.getvar('val%d' % i, HyphyInterface.STRING) for i in range(numjobs) ]


class HyphyMap(object):

    def __init__(self, batchfile, retvar):
        if not exists(batchfile):
            raise ValueError('need to provide a real template batchfile!')
        self._batchfile = abspath(batchfile)
        self._retvar = retvar
        nodes = mpi_node_count()
        if nodes > 0:
            self._mpi = True
            self._nodes = nodes
        else:
            self._mpi = False
            self._nodes = cpu_count()

    @property
    def nodes(self):
        return self._nodes

    def map(self, argslist, globalvars={}, quiet=True):
        numjobs = len(argslist)
        if self._mpi:
            # message passing interface
            # don't fuck with the number of backslashes.
            # Seriously. You don't know what you're doing.
            cmds = dedent('''\
            GLOBAL_FPRINTF_REDIRECT = "/dev/null";
            _job = 0;
            %(jobopts)s
            _jobvals = {};
            _nodestates = { MPI_NODE_COUNT-1, 2 };
            _received = 0;
            while ( _received < %(numjobs)d ) {
                _recvjob = 1;
                if ( _job < %(numjobs)d ) {
                    for ( _node = 0; _node < MPI_NODE_COUNT-1; _node += 1 ) {
                        if ( _nodestates[ _node ][ 0 ] == 0 ) {
                            _recvjob = 0;
                            break;
                        }
                    }
                    if ( ! _recvjob ) {
                        _mpicmds = "";
                        _mpicmds * 256;
                        _mpicmds * ( "%(globalvars)s" );
                        _mpicmds * ( "MESSAGE_LOGGING = 0;" );
                        _mpicmds * ( "_options = " + _jobopts[ _job ] + ";" );
                        _mpicmds * ( "ExecuteAFile( "{1}%(batchfile)s"{1}, _options );" );
                        _mpicmds * ( "_retstr = "{1}"{1};" );
                        _mpicmds * ( "_retstr * 128;" );
                        _mpicmds * ( "_retstr * ( "{1}_retjob = "{1} + " + _job + " + "{1};"{1} );" );
                        _mpicmds * ( "_retstr * ( "{1}_retval = "{2}"{1} + ( %(retvar)s ^ {{ "{1}"{2}"{1}, "{1}"{3}"{1} }} ) + "{1}"{2};"{1} );" );
                        _mpicmds * ( "_retstr * 0;" );
                        _mpicmds * ( "return _retstr;" );
                        _mpicmds * 0;
                        MPISend( _node + 1, _mpicmds );
                        _nodestates[ _node ][ 0 ] = 1;
                        _nodestates[ _node ][ 1 ] = Time(0);
                        _job += 1;
                    }
                }
                if ( _recvjob ) {
                    MPIReceive( -1, _null, _retstr );
                    ExecuteCommands( _retstr );
                    _jobvals[ _retjob ] = _retval;
                    _received += 1;
                }
            }
            GLOBAL_FPRINTF_REDIRECT = "";
            fprintf( stdout, "[" + _jobvals[ 0 ] );
            for ( _job = 1; _job < %(numjobs)d; _job += 1 ) {
                fprintf ( stdout, "," + _jobvals[ _job ] );
            }
            fprintf( stdout, "]" );
            ''') % {
                'batchfile': self._batchfile,
                'globalvars': _globalvars(globalvars).replace('\n', '\\n').replace('"', '\\"'),
                'numjobs': numjobs,
                'retvar': self._retvar,
                'jobopts': _jobopts(argslist),
            }

            # this facility helps us write code that doesn't break
            # by allowing us to escape quotations appropriately:
            # by depth
            digits = set()
            for m in finditer(r'"{(\d+)}', cmds):
                digits.add(int(m.group(1)))

            for nslash in digits:
                pwr = 2 ** nslash - 1
                cmds = cmds.replace('"{%d}' % nslash, ('\\' * pwr) + '"')

            retcode, pout, perr = _runhyphympi(cmds)

            # the following no longer makes sense given the child process nature
            # of how we call hyphympi

            # if not quiet:
            #     if pout != '':
            #         print(pout, file=stderr)

            if perr != '':
                raise RuntimeError(perr)

            # this is a hideous parser for the outermost
            # json-esque array that the script above outputs

            # start at 1 because the first char is a '['
            i = 1
            nb = 0
            retarr = []
            for j, c in enumerate(pout):
                # every time we encounter a '[', increment nb
                if c == '[':
                    nb += 1
                # likewise, if we encounter a closing bracket, decrement
                if c == ']':
                    nb -= 1
                # if we're at a comma and nb == 1, we're in the outer list
                # so split from the last time we split to j-1
                if c == ',' and nb == 1:
                    retarr.append(pout[i:j])
                    i = j + 1

            # don't forget the final piece
            retarr.append(pout[i:(j + nb)])

            return retarr
        else:
            # multiprocessing
            numjobs = len(argslist)
            globals = _globalvars(globalvars)
            results = farmout(
                num=numjobs,
                setup=lambda i: (
                    _jobdispatch,
                    self._batchfile,
                    self._retvar,
                    globals,
                    (argslist[i],),
                    quiet
                ),
                worker=farmworker,
                isresult=lambda r: isinstance(r, list),
                attempts=3
            )
            return list(chain(*results))




from __future__ import division, print_function

from copy import deepcopy
from os import getcwd
from os.path import abspath, exists, join

import HyPhy as hp


__all__ = ['HyphyInterface', 'escape', 'tohyphy']


def escape(value):
    if not isinstance(value, (float, int, str)):
        raise ValueError("Cannot escape types other than float, int, or str: '%s'" % repr(value))
    return '"%s"' % value.replace('"', r'\"') if isinstance(value, str) else repr(value)


def tohyphy(name, value):
    execstr = ''
    if hasattr(value, '_tohyphy'):
        execstr += value._tohyphy(name)
    elif isinstance(value, (list, tuple)):
        for i, v in enumerate(value):
            if isinstance(v, (list, tuple)):
                if i == 0:
                    execstr += "%s = {%d,%d};\n" % (name, len(value), len(v))
                for j, w in enumerate(v):
                    if not isinstance(w, (float, int)):
                        raise ValueError("2D matrices must contain values of type float or int")
                    execstr += "%s[%d][%d] = %s;\n" % (name, i, j, repr(w))
            elif isinstance(v, (float, int, str)):
                if i == 0:
                    execstr += "%s = {};\n" % name
                execstr += "%s[%d] = %s;\n" % (name, i, escape(v))
            else:
                raise ValueError("Lists must contain values of type float, int, or str")
    elif isinstance(value, dict):
        initted = False
        for k, v in value.items():
            if not initted:
                execstr += "%s = {};\n" % name
                initted = True
            if not isinstance(k, (int, str)):
                raise ValueError("Dictionary keys must be values of type int or str")
            if not isinstance(v, (float, int, str)):
                raise ValueError("Dictionaries must contain values of type float, int, or str")
            execstr += "%s[%s] = %s;\n" % (name, escape(k), escape(v))
    elif isinstance(value, (float, int, str)):
        execstr += "%s = %s;\n" % (name, escape(value))
    else:
        raise ValueError("inject() supports only floats, ints, and strs; lists of the same; and 2d matrices of floats or ints")
    return execstr


class HyphyInterface(object):

    MATRIX = hp.THYPHY_TYPE_MATRIX
    NUMBER = hp.THYPHY_TYPE_NUMBER
    STRING = hp.THYPHY_TYPE_STRING

    def __init__(self, batchfile=None, num_cpus=1):
        self._batchfile = batchfile
        self._execstr = ''
        self._instance = hp._THyPhy(join(abspath(hp.__file__), 'res'), num_cpus)
        self._stdout = ''
        self._stderr = ''
        self._warnings = ''

    def _fetchenv(self):
        self._stdout += self._instance.GetStdout().sData.strip()
        self._stderr += self._instance.GetErrors().sData.strip()
        self._warnings += self._instance.GetWarnings().sData.strip()

    def getvar(self, variable, typ):
        _res = self._instance.AskFor(variable)
        if typ not in (HyphyInterface.MATRIX, HyphyInterface.NUMBER, HyphyInterface.STRING):
            raise ValueError("Unknown type supplied: please use one of HyphyInterface.{MATRIX,NUMBER,STRING}")
        if (self._instance.CanCast(_res, typ)):
            res = self._instance.CastResult(_res, typ)
            if typ == HyphyInterface.STRING:
                return res.castToString().sData
            elif typ == HyphyInterface.NUMBER:
                return res.castToNumber().nValue
            elif typ == HyphyInterface.MATRIX:
                hymat = res.castToMatrix()
                mat = None
                one_d = hymat.mRows == 1, hymat.mCols == 1
                if any(one_d):
                    dim = max(hymat.mRows, hymat.mCols)
                    mat = [0.] * dim
                    if one_d[0]:
                        for i in range(dim):
                            mat[i] = hymat.MatrixCell(0, i)
                    else:
                        for i in range(dim):
                            mat[i] = hymat.MatrixCell(i, 0)
                else:
                    mat = [[None] * hymat.mCols] * hymat.mRows
                    for i in range(hymat.mRows):
                        for j in range(hymat.mCols):
                            mat[i][j] = hymat.MatrixCell(i, j)
                return mat
            else:
                # dead code, we assume
                assert(0)
        else:
            raise RuntimeError("Cast failed in HyphyInterface, assume an incorrect type was supplied for variable `%s'" % variable)

    def queuecmd(self, execstr):
        self._execstr += execstr

    def queuestralloc(self, name, size):
        self._execstr += '%s = "";\n%s * %d;\n' % (name, name, size)

    def queuevar(self, name, value):
        self._execstr += tohyphy(name, value)

    def reset(self):
        # XXX: clear self._execstr???
        self._instance.ClearAll()
        self._stderr = ''
        self._stdout = ''
        self._warnings = ''

    def runqueue(self, *args, **kwargs):
        batchfile = kwargs.pop('batchfile', None)
        execstr = kwargs.pop('execstr', None)

        errstr = "runqueue() takes a two optional arguments: a HyPhy 'batchfile' or an 'execstr' containing HyPhy commands"

        # if not already provided by kwargs,
        # pop vars from args (if they're there)
        if execstr is None:
            if len(args) > 1:
                execstr = args.pop(1)
        if batchfile is None:
            if len(args) > 0 and exists(args[0]):
                batchfile = args.pop(0)

        if len(args) > 0 or len(kwargs) > 0:
            raise ValueError(errstr)

        # if the execstr wasn't provided,
        # grab the instance one
        if execstr is None:
            execstr = self._execstr

        # if we weren't given a batchfile in runqueue,
        # grab the instance one
        if batchfile is None:
            batchfile = self._batchfile

        if batchfile is not None:
            if exists(batchfile):
                with open(batchfile) as fh:
                    execstr += fh.read()
            else:
                raise ValueError("Invalid batchfile `%s', it doesn't exist!" % batchfile)

        ret = self._instance.ExecuteBF(execstr)
        HyphyInterface._fetchenv(self)
        return ret

    @property
    def stderr(self):
        return deepcopy(self._stderr)

    @property
    def stdout(self):
        return deepcopy(self._stdout)

    @property
    def warnings(self):
        return deepcopy(self._warnings)



#!/usr/bin/env python3

from __future__ import division, print_function

import sys

from os.path import abspath, join, split
from setuptools import setup

sys.path.insert(0, join(split(abspath(__file__))[0], 'lib'))

setup(name='hyphy-helper',
      version='0.9.7',
      description='An intuitive HyPhy interface for Python',
      author='N Lance Hepler',
      author_email='nlhepler@gmail.com',
      url='http://github.com/veg/hppy',
      license='GNU GPL version 3',
      packages=['hppy'],
      package_dir={'': 'lib'},
      install_requires=[
          'HyPhy >=0.1.1',
          'Cython >=0.22.1',
          'fakemp >= 0.9.1'
      ]
     )
